package uulm.teamname.marvelous.server.netconnector;

import org.java_websocket.handshake.ClientHandshake;
import org.tinylog.Logger;
import uulm.teamname.marvelous.gamelibrary.json.JSON;
import uulm.teamname.marvelous.gamelibrary.messages.BasicMessage;
import uulm.teamname.marvelous.gamelibrary.messages.ErrorMessage;
import uulm.teamname.marvelous.gamelibrary.messages.client.*;
import uulm.teamname.marvelous.server.lobbymanager.Participant;

import org.java_websocket.WebSocket;

import java.util.*;

public class UserManager {

    /** A set of users that aren't assigned to lobbies or character selection yet */
    private final HashSet<WebSocket> newUsers;

    /** A set of users that can reconnect if they wish to do so, and their matching Participants */
    private final HashMap<WebSocket, Participant> readyToReconnect;

    /** A set of users that only have to send the
     * {@link uulm.teamname.marvelous.gamelibrary.messages.client.PlayerReadyMessage} to be assigned
     * to a lobby, containing WebSockets mapped to their user's usernames
     */
    private final HashMap<WebSocket, SUID> readyToConnect;


    /** A set of users that are already assigned to lobbies or character selection, mapped to their participants */
    private final HashMap<WebSocket, Participant> inGame;

    /**
     * A map mapping {@link SUID SUIDs} to {@link Participant Participants} to assert whether (and where to) reconnect
     * an user
     */
    private final HashMap<SUID, Participant> activeParticipants;

    private final JSON json;

    /** Constructs a new, empty UserManager */
    public UserManager(JSON json) {
        this.newUsers = new HashSet<>();
        this.readyToConnect = new HashMap<>();
        this.readyToReconnect = new HashMap<>();
        this.inGame = new HashMap<>();
        this.activeParticipants = new HashMap<>();
        this.json = json;
    }

    /** Called on a new WebSocket connection. Places the WebSocket and its ResourceDescriptor in a HashMap. */
    public void connectUser(WebSocket conn, ClientHandshake handshake) {
        synchronized (newUsers) {
            newUsers.add(conn);
        }
    }

    public void messageReceived(WebSocket conn, String message) {
        // TODO: Implement more than only a basic handshake

        var parsedMessageOptional = json.parse(message);
        if (parsedMessageOptional.isEmpty()) {
            Logger.debug("Invalid message '{}' received, sending error", message);
            sendError(conn, "Message could not be parsed");
        } else {
            var parsedMessage = parsedMessageOptional.get();

            if (parsedMessage instanceof HelloServerMessage)
                 handshake(conn, (HelloServerMessage) parsedMessage);
        }
    }

    void handshake(WebSocket conn, HelloServerMessage message) {
        if (!newUsers.contains(conn)) {
            Logger.debug("websocket {} sent HelloServerMessage outside of handshake", conn);
            sendError(conn, "Invalid message, as Handshake is already completed");
            return;
        }

        Logger.info("Performing handshake with user '{}'", message.name);

        SUID clientID = new SUID(message.name, message.deviceID);

        // check if client is reconnected
        var participant = activeParticipants.get(clientID);
        if (participant != null) {
            synchronized (readyToReconnect) {
                readyToReconnect.put(conn, participant);
            }
        } else {
            Logger.trace("removing handshaking user from newUsers");
            synchronized (newUsers) {
                newUsers.remove(conn);
            }
            Logger.trace("adding handshaking user to readyToConnect");
            synchronized (readyToConnect) {
                readyToConnect.put(conn, clientID);
            }
        }
    }

    public void reconnectClient(WebSocket conn, ReconnectMessage message) {
        if (!readyToConnect.containsKey(conn)) {
            Logger.debug("Non-reconnect-allowed client has sent reconnect message, sending error");
            sendError(conn, "Reconnect is not possible");
            return;
        }
        Logger.info("Reconnecting client {} to their lobby");
        var participantToRestore = readyToReconnect.get(conn);
        synchronized (participantToRestore) {
            participantToRestore.setConnection(conn);
        }
        synchronized (readyToReconnect) {
            readyToReconnect.remove(conn);
        }
        synchronized (inGame) {
            inGame.put(conn, participantToRestore);
        }
        // activeParticipants remains the same, as no players have been removed from the game
    }

    public void assignLobby(WebSocket conn, PlayerReadyMessage message) {
    }

    public void charactersSelected(WebSocket conn, CharacterSelectionMessage message) {
    }

    public void relayRequestMessage(Participant conn, RequestMessage message) {
    }

    /** Sends an {@link uulm.teamname.marvelous.gamelibrary.messages.ErrorMessage} to the specified user. */
    public void sendError(WebSocket conn, String error) {
        Logger.debug("Sending error message '{}' to WebSocket {}", error, conn);
        var errorMessage = new ErrorMessage();
        errorMessage.message = error;
        conn.send(json.stringify(errorMessage).get());
    }

    public void disconnectUser(WebSocket conn, boolean closedByRemote) {
        // TODO: notify clients and such if the connection was in fact closed by the remote. Also remove participant
        synchronized (newUsers) {
            newUsers.remove(conn);
        }
        synchronized (inGame) {
            inGame.remove(conn);
        }
    }

    public boolean isUserConnected(WebSocket user) {
        return newUsers.contains(user) ||
                readyToReconnect.containsKey(user) ||
                readyToConnect.containsKey(user) ||
                inGame.containsKey(user);
    }

    public int getUserCount() {
        return newUsers.size() + readyToConnect.size() + readyToReconnect.size() + inGame.size();
    }

    /** Package-private getter for mutable newUsers HashSet, meant for testing */
    HashSet<WebSocket> getNewUsers() {
        return newUsers;
    }

    /** Package-private getter for mutable readyToReconnect HashMap, meant for testing */
    HashMap<WebSocket, Participant> getReadyToReconnect() {
        return readyToReconnect;
    }

    /** Package-private getter for mutable readyToConnect HashMap, meant for testing */
    HashMap<WebSocket, SUID> getReadyToConnect() {
        return readyToConnect;
    }

    /** Package-private getter for mutable inGame HashMap, meant for testing */
    HashMap<WebSocket, Participant> getInGame() {
        return inGame;
    }

    /** Package-private getter for mutable activeParticipants HashMap, meant for testing */
    HashMap<SUID, Participant> getActiveParticipants() {
        return activeParticipants;
    }
}
